import java.io.File;
import java.io.FileNotFoundException;
import java.util.Iterator;
import java.util.Map;
import java.util.Scanner;
import java.util.TreeMap;

public class HuffmanGettysburg
{
    /**
     * Build and return a frequency map of the characters
     * in an input text file. The keys are characters,
     * and the values are the frequencies of the characteres.
     * @param in input text file Scanner
     * @return the frequency map
     */
    private static Map<Character, Integer> makeFrequencyMap(Scanner in)
    {
        Map<Character, Integer> map = new TreeMap<Character, Integer>();
        in.useDelimiter("");
        
        while (in.hasNextLine()) {
            char ch = in.next().charAt(0);
            Integer frequency = map.get(ch);
            
            if (frequency == null) frequency = 1;
            else                   frequency++;
            
            map.put(ch, frequency);
        }
        
        return map;
    }
    
    /**
     * Compute the total number of characters in the input text file.
     * @param map the frequency map
     * @return the total number of characters
     */
    private static int computeCharacterCount(Map<Character, Integer> map)
    {
        int count = 0;
        Iterator<Character> it = map.keySet().iterator();
        
        while (it.hasNext()) {
            char ch = it.next();
            int frequency = map.get(ch);
            count += frequency;
        }
        
        return count;
    }
    
    /**
     * Print the Huffman codes of the characters as a table.
     * Print the characters in alphabetical order and their frequencies
     * and Huffman codes, one character per row.
     * @param frequencyMap
     * @param encodingMap
     * @return the total length of all the Huffman codes
     */
    private static int printHuffmanCodes(Map<Character, Integer> frequencyMap,
                                         Map<Character, String>  encodingMap)
    {
        int length = 0;
        Iterator<Character> it = frequencyMap.keySet().iterator();
        
        System.out.println("\nCharacter Frequency Code");
        
        while (it.hasNext()) {
            char ch = it.next();
            int frequency = frequencyMap.get(ch);
            String str = new String();
            
            switch (ch) {
                case ' '  : str += "space"; break;
                case '\n' : str += "\\n";   break;
                default:    str += ch;
            }
            
            String code = encodingMap.get(ch);
            length += frequency*code.length();
            
            System.out.printf("%6s %8d     %s\n", str, frequency, code);
        }
        
        return length;
    }
    
    private static final String INPUT_FILE = "GettysburgAddress.txt";
    
    public static void main(String[] args)
    {
        Scanner in = null;
        
        try {
            in = new Scanner(new File(INPUT_FILE));
            Map<Character, Integer> frequencyMap = makeFrequencyMap(in);
            
            int count = computeCharacterCount(frequencyMap);
            int utfBits = 8*count;
            System.out.printf("%5d total characters\n", count);
            System.out.printf("%5d total UTF-8 bits\n", utfBits);
            
            HuffmanTree tree = new HuffmanTree(frequencyMap);
            Map<Character, String> encodingMap = tree.getEncodingMap();
            
            int encodedBits = printHuffmanCodes(frequencyMap, encodingMap);
            System.out.printf("\n%5d total Huffman-encoded bits\n", 
                              encodedBits);
            
            double reduction = 
                    100.0*(utfBits - encodedBits)/((double) utfBits);
            System.out.printf("%4.1f%% reduction", reduction);
        }
        catch (FileNotFoundException ex) {
            System.out.printf("*** ERROR: File \"%s\" not found.\n",
                              INPUT_FILE);
        }
        finally {
            if (in != null) in.close();
        }
    }
}
